/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "turn.h"

pokerturn_t pokerTurnGet(poker_t *poker, uint8_t playerIndex) {
  int32_t random, maxBet, bluffBet, callBet;
  float winProbability, expectedGain, confidence, potOdds;
  bool isBluff;
  int32_t amount;
  pokerplayer_t *player;
  pokerplayerwinning_t winning;
  uint8_t i, cardNumber0, cardNumber1, suitNumber0, suitNumber1;

  player = poker->players + playerIndex;

  // Can the player do anything?
  if(player->state & (POKER_PLAYER_STATE_FOLDED | POKER_PLAYER_STATE_OUT)) {
    return pokerTurnOut(poker, playerIndex);
  }

  // Now we basically need to determine the AI's action here. The AI really only
  // needs to do one thing, decide whether or not they want to Bet or Not.
  // In future I may make checking optional, but for now there's really no 
  // reason not to.
  pokerWinnerPlayerGet(&poker->dealer, player, &winning);

  // Get the current winning hand.
  if(poker->dealer.cardsFacing == 0 && player->cardCount >= 2) {
    cardNumber0 = cardGetNumber(player->cards[0]);
    cardNumber1 = cardGetNumber(player->cards[1]);
    suitNumber0 = cardGetSuit(player->cards[0]);
    suitNumber1 = cardGetSuit(player->cards[1]);

    i = mathAbs(cardNumber0 - cardNumber1);

    confidence = (float)cardNumber0 + (float)cardNumber1;
    if(cardNumber0 == cardNumber1) {
      confidence += 6;
    } else if(suitNumber0 == suitNumber1) {
      confidence += 4;
    }

    if(i > 4) {
      confidence -= 4;
    } else if(i > 2) {
      confidence -= i;
    }

    confidence = confidence / 30.0f;
  } else {
    confidence = pokerWinnerGetTypeConfidence(winning.type);
  }

  callBet = poker->bet.currentBet - player->currentBet;
  winProbability = confidence;

  if(callBet > 0) {
    potOdds = (float)callBet / ((float)callBet + (float)poker->bet.pot);
  } else {
    potOdds = (
      1.0f / (float)pokerBetGetRemainingPlayerCount(&poker->bet, poker->players)
    );
  }
  expectedGain = winProbability / potOdds;

  random = randInt32() % 100;
  maxBet = (int32_t)(poker->bet.currentBet / 1.75f) - (random / 2);//?
  maxBet -= callBet;//?
  bluffBet = random * maxBet / 100 / 2;

  isBluff = false;
  amount = 0;

  if(expectedGain < 0.8 && winProbability < 0.8) {
    if(random < 95) {
      amount = 0;//FOLD
    } else {
      amount = bluffBet;
      isBluff = true;
    }
  } else if ((expectedGain < 1.0 && winProbability < 0.85) || winProbability < 0.1) {
    if (random < 80) {
      amount = 0;//FOLD
    } else if(random < 5) {
      amount = callBet;
      isBluff = true;
    } else {
      amount = bluffBet;
      isBluff = true;
    }
  } else if ((expectedGain < 1.3 && winProbability < 0.9) || winProbability < 0.5) {
    if (random < 60 || winProbability < 0.5) {
      amount = callBet;
    } else {
      amount = maxBet;
    }
  } else if (winProbability < 0.95 || poker->dealer.cardsFacing < 0x04) {
    if(random < 30) {
      amount = callBet;
    } else {
      amount = maxBet;
    }
  } else {
    amount = (poker->bet.currentBet - callBet) * 9 / 10;
  }
  

  printf("Raw Amount %i\n", amount);

  // TODO: Make sure we don't get stuck in a raise loop.
  if(amount > 0) {
    amount = mathMin(amount, callBet);
    return pokerTurnRaise(poker, playerIndex, amount);
  }
  
  if(pokerTurnCanPlayerCheck(poker, playerIndex)) {
    return pokerTurnCheck(poker, playerIndex);
  }

  // Nothing worth doing, so let's just fold here.
  return pokerTurnFold(poker, playerIndex);
}

void pokerTurnAction(poker_t *poker, pokerplayer_t *player, pokerturn_t *turn) {
  switch(turn->type) {
    case POKER_TURN_TYPE_BET:
    case POKER_TURN_TYPE_CALL:
    case POKER_TURN_TYPE_ALL_IN:
      pokerBetPlayer(&poker->bet, player, turn->chips);
      break;

    case POKER_TURN_TYPE_CHECK:
      pokerBetPlayer(&poker->bet, player, 0);
      break;

    case POKER_TURN_TYPE_FOLD:
      player->state |= POKER_PLAYER_STATE_FOLDED;
      break;
  }
  
  player->state |= POKER_PLAYER_STATE_ROUND_MOVE;
}

pokerturn_t pokerTurnOut(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_OUT;
  return turn;
}

pokerturn_t pokerTurnFold(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_FOLD;
  return turn;
}

pokerturn_t pokerTurnCheck(poker_t *poker, uint8_t player) {
  pokerturn_t turn;
  turn.type = POKER_TURN_TYPE_CHECK;
  return turn;
}

pokerturn_t pokerTurnCall(poker_t *poker, uint8_t playerIndex) {
  pokerturn_t turn;
  pokerplayer_t *player;
  player = poker->players + playerIndex;
  turn = pokerTurnRaise(
    poker, playerIndex, poker->bet.currentBet - player->currentBet
  );
  return turn;
}

pokerturn_t pokerTurnRaise(poker_t *poker, uint8_t playerIndex, int32_t chips) {
  pokerturn_t turn;
  pokerplayer_t *player;
  player = poker->players + playerIndex;

  if(chips < poker->bet.currentBet) chips = poker->bet.currentBet;

  if(player->chips <= chips) {
    turn.chips = player->chips;
    turn.type = POKER_TURN_TYPE_ALL_IN;
  } else {
    turn.chips = chips;
    turn.type = POKER_TURN_TYPE_BET;

    if(chips == (poker->bet.currentBet - player->currentBet)) {
      turn.type = POKER_TURN_TYPE_CALL;
    }
  }

  return turn;
}

bool pokerTurnCanPlayerCheck(poker_t *poker, uint8_t playerIndex) {
  return (poker->players + playerIndex)->currentBet >= poker->bet.currentBet;
}